#include <iostream>
#include <stdio.h>

#include "system/memory.h"				//make this an object?
#include "system/singletons.h"
#include "osxsystem/singletons.h"

#ifdef COCOA_INSTEAD_OF_CARBON
extern "C" {
	int cocoaCreateWindow(float, float, float, float);
	int cocoaShutdown();
	WindowRef cocoaGetWindowRef();
}
#endif

using namespace std;

//////////////////////////////// key map / mouse map

static Key::Key_t keyForVirtKey[256];

//all i have is my apple keyboard to go off of =)
static void initKeyMaps() {
	memset(keyForVirtKey, 0, sizeof(keyForVirtKey));
#define VKBRIDGE(x,y)	keyForVirtKey[x] = (Key::Key_t)y;
	VKBRIDGE(53, Key::ESC)
	VKBRIDGE(122, Key::F1)
	VKBRIDGE(120, Key::F2)
	VKBRIDGE(99, Key::F3)
	VKBRIDGE(118, Key::F4)
	VKBRIDGE(96, Key::F5)
	VKBRIDGE(97, Key::F6)
	VKBRIDGE(98, Key::F7)
	VKBRIDGE(100, Key::F8)
	//f9-f11 are currently finder-bound so i can't see them atm
	VKBRIDGE(111, Key::F12)
	//eject?

	VKBRIDGE(50, '`')
	VKBRIDGE(18, '1')
	VKBRIDGE(19, '2')
	VKBRIDGE(20, '3')
	VKBRIDGE(21, '4')
	VKBRIDGE(23, '5')
	VKBRIDGE(22, '6')
	VKBRIDGE(26, '7')
	VKBRIDGE(28, '8')
	VKBRIDGE(25, '9')
	VKBRIDGE(29, '0')
	VKBRIDGE(27, '-')
	VKBRIDGE(24, '=')
	VKBRIDGE(51, Key::BACKSPACE)	//it says 'delete' but it works like 'backspace'

	VKBRIDGE(48, Key::TAB)
	VKBRIDGE(12, 'q')
	VKBRIDGE(13, 'w')
	VKBRIDGE(14, 'e')
	VKBRIDGE(15, 'r')
	VKBRIDGE(17, 't')
	VKBRIDGE(16, 'y')
	VKBRIDGE(32, 'u')
	VKBRIDGE(34, 'i')
	VKBRIDGE(31, 'o')
	VKBRIDGE(35, 'p')
	VKBRIDGE(33, '[')
	VKBRIDGE(30, ']')
	VKBRIDGE(42, '\\')

	VKBRIDGE(0, 'a')
	VKBRIDGE(1, 's')
	VKBRIDGE(2, 'd')
	VKBRIDGE(3, 'f')
	VKBRIDGE(5, 'g')
	VKBRIDGE(4, 'h')
	VKBRIDGE(38, 'j')
	VKBRIDGE(40, 'k')
	VKBRIDGE(37, 'l')
	VKBRIDGE(41, ';')
	VKBRIDGE(39, '\'')
	VKBRIDGE(36, Key::RETURN)

	VKBRIDGE(6, 'z')
	VKBRIDGE(7, 'x')
	VKBRIDGE(8, 'c')
	VKBRIDGE(9, 'v')
	VKBRIDGE(11, 'b')
	VKBRIDGE(45, 'n')
	VKBRIDGE(46, 'm')
	VKBRIDGE(43, ',')
	VKBRIDGE(47, '.')
	VKBRIDGE(44, '/')
	
	VKBRIDGE(49, Key::SPACE)
	VKBRIDGE(76, Key::RETURN)	//technically enter
	VKBRIDGE(126, Key::UP)
	VKBRIDGE(125, Key::DOWN)
	VKBRIDGE(123, Key::LEFT)
	VKBRIDGE(124, Key::RIGHT)
}

//////////////////////////////// static connection code

OSStatus MouseEventHandler(EventHandlerCallRef inHandlerCallRef, EventRef inEvent, void *inUserData) {
	UInt32 buttonState = GetCurrentEventButtonState();
	for (int i = 0; i < mouse.MAX_BUTTON; i++) {	//convenient 1-1 mapping here between osx and my stuff
		mouse.setButtonPress((MouseHandler::Button_t)i,  (buttonState >> i) & 1);
	}
	
	//todo - if keystate LMETA or RMETA is down and mouse1 is pressed then turn it into a mouse2 press ...
	//messy ... and i was hoping that the os handled that sort of thing
	if (kbd.keystate[Key::LMETA] || kbd.keystate[Key::RMETA]) {
		mouse.setButtonPress(MouseHandler::BUTTON_LEFT, 0);
		mouse.setButtonPress(MouseHandler::BUTTON_RIGHT, buttonState & 1);
	}
	
	HIPoint windowMouseLocation;
	GetEventParameter(inEvent, kEventParamWindowMouseLocation, typeHIPoint, NULL, sizeof(HIPoint), NULL, &windowMouseLocation);
	//TODO - account for the titlebar in the window coordinates (to view coords?)
	mouse.setCursorPos(
		(int)windowMouseLocation.x,
		(int)windowMouseLocation.y - 24);

	return CallNextEventHandler(inHandlerCallRef, inEvent);
}

OSStatus MouseWheelEventHandler(EventHandlerCallRef inHandlerCallRef, EventRef inEvent, void *inUserData) {
	int axis = 0;
	GetEventParameter(inEvent, kEventParamMouseWheelAxis, typeMouseWheelAxis, NULL, sizeof(axis), NULL, &axis);
	
	int delta = 0;	//turns out "typeLongInteger" is for signed 32-bit integers ... just like "typeInteger"
	GetEventParameter(inEvent, kEventParamMouseWheelDelta, typeLongInteger, NULL, sizeof(delta), NULL, &delta);
	
	if (axis == 1) {	//the y axis is the default wheel button ... 
		mouse.accumDZ(delta);
	}
	
	return 0;
}

OSStatus KeyUpEventHandler(EventHandlerCallRef inHandlerCallRef, EventRef inEvent, void *inUserData) {
	UInt32 keycode;
	GetEventParameter(inEvent, kEventParamKeyCode, typeUInt32, NULL, sizeof(UInt32), NULL, &keycode);
	kbd.keyUp(keyForVirtKey[keycode]);

	return CallNextEventHandler(inHandlerCallRef, inEvent);
}

OSStatus KeyDownEventHandler(EventHandlerCallRef inHandlerCallRef, EventRef inEvent, void *inUserData) {
	UInt32 keycode;
	GetEventParameter(inEvent, kEventParamKeyCode, typeUInt32, NULL, sizeof(UInt32), NULL, &keycode);
	kbd.keyDown(keyForVirtKey[keycode]);

	return CallNextEventHandler(inHandlerCallRef, inEvent);
}

OSStatus KeyboardModifierEventHandler(EventHandlerCallRef inHandlerCallRef, EventRef inEvent, void *inUserData) {
	UInt32 modifiers;
	GetEventParameter(inEvent, kEventParamKeyModifiers, typeUInt32, NULL, sizeof(UInt32), NULL, &modifiers);
	
	//so far i don't know how to distinctify left & right ...
	bool cmd = !!(modifiers & (1 << 8));		//apple. meta?
	bool shift = !!(modifiers & (1 << 9));
	bool option = !!(modifiers & (1 << 11));	//alt
	bool ctrl = !!(modifiers & (1 << 12));
	bool rightShift = !!(modifiers & (1 << 13));	//i hear these dont always work ...
	bool rightOption = !!(modifiers & (1 << 14));
	bool rightCtrl = !!(modifiers & (1 << 15));

#define VKMODIFIERBRIDGE(x,y)	if (x) kbd.keyDown(Key::y); else kbd.keyUp(Key::y);
	VKMODIFIERBRIDGE(shift, LSHIFT)
	VKMODIFIERBRIDGE(rightShift, RSHIFT)
	VKMODIFIERBRIDGE(ctrl, LCTRL)
	VKMODIFIERBRIDGE(rightCtrl, RCTRL)
	VKMODIFIERBRIDGE(option, LALT)
	VKMODIFIERBRIDGE(rightOption, RALT)
	VKMODIFIERBRIDGE(cmd, LMETA)
	VKMODIFIERBRIDGE(cmd, RMETA)
	
	return CallNextEventHandler(inHandlerCallRef, inEvent);
}

OSStatus WindowEventHandler(EventHandlerCallRef inHandlerCallRef, EventRef inEvent, void *inUserData) {
	UInt32 eventKind = GetEventKind(inEvent);
//	printf("WindowEventHandler %d\n", eventKind);
	switch (eventKind) {
	case kEventWindowClose:
		sys.requestExit();
		break;
	case kEventWindowBoundsChanged:
		{
			Rect rect;
			GetWindowBounds(osxSys.getWnd(), kWindowContentRgn, &rect);
			vec2i newsize(rect.right - rect.left, rect.bottom - rect.top);
			disp.resize(VEC2ELEM(newsize));
			//cout << "disp size " << newsize << endl;
#ifdef COCOA_INSTEAD_OF_CARBON
			CGLSetCurrentContext(osxSys.getCglCtx());
#else
			aglSetDrawable(osxSys.getAglCtx(), (AGLDrawable)GetWindowPort(osxSys.getWnd()));
#endif
		}
		break;
	}
	
	return CallNextEventHandler(inHandlerCallRef, inEvent);
}

//////////////////////////////// class code

void OSXSystem::reset() {
	//init data values here
	wnd = NULL;
#ifdef COCOA_INSTEAD_OF_CARBON
	cglCtx = NULL;
#else
	aglCtx = NULL;
#endif
	exitRequested = false;
	initDoubleBuffered = doubleBuffered = true;

	loopCallback = NULL;
	initFullScreen = false;
	
	vec2i screenSize(1024,768);
	initSize.set(720,480);
	initPos = screenSize/2 - initSize/2;
}

void OSXSystem::init() {
	//hmm... i'd like to expose this as a parameter
	//but how to make the parameter set arbitrary across various systems?
	doubleBuffered = initDoubleBuffered;
	
	timer.init();
	memSystemInit();
	
	//init our windows
	
//	IBNibRef        nibRef;
//	OSStatus err;
//	err = CreateNibReference(CFSTR("main"), &nibRef);
//	cout << "CreateNibReference " << err << endl;
//	err = SetMenuBarFromNib(nibRef, CFSTR("MenuBar"));
//	cout << "SetMenuBarFromNib " << err << endl;
//	DisposeNibReference(nibRef);

	
	//window 
	{
#ifdef COCOA_INSTEAD_OF_CARBON
		cocoaCreateWindow(initPos.x, initPos.y, initSize.x, initSize.y);
		wnd = cocoaGetWindowRef();
		//TODO - find how to embed a .nib into a .so ... yup
#else	
		//TODO - create a titlebar
		

		//WindowGroupRef group;
		//CreateWindowGroup(0, &group);

		Rect rect;
		rect.left = initPos.x;
		rect.top = initPos.y;
		rect.right = initPos.x + initSize.x;
		rect.bottom = initPos.y + initSize.y;
		CreateNewWindow(kDocumentWindowClass, 
			//kWindowStandardDocumentAttributes |
				kWindowCloseBoxAttribute |
				kWindowFullZoomAttribute |
				kWindowCollapseBoxAttribute |
				kWindowResizableAttribute |
			kWindowStandardHandlerAttribute |
			kWindowInWindowMenuAttribute |
			kWindowLiveResizeAttribute
		, &rect, &wnd);
		
//		RepositionWindow(wnd, NULL, kWindowCascadeOnMainScreen );
//		SetWindowModified( wnd, false );
//		MacShowWindow( wnd );
		ShowWindow(wnd);
		SetPortWindowPort( wnd );
//		BringToFront(wnd);
	}
	
	//opengl context
	{
		//seems like with wgl there were ways to negotiate your pixel format
		//what you wanted vs what was available
		GLint attribs[32];
		GLint *ap = attribs;
		*(ap++) = AGL_RGBA;
		*(ap++) = AGL_RED_SIZE;	*(ap++) = 8;
		*(ap++) = AGL_GREEN_SIZE; *(ap++) = 8;
		*(ap++) = AGL_BLUE_SIZE; *(ap++) = 8;
		*(ap++) = AGL_ALPHA_SIZE; *(ap++) = 8;
		*(ap++) = AGL_DEPTH_SIZE; *(ap++) = 32;
		//*(ap++) = AGL_STENCIL_SIZE; *(ap++) = 16;	//messes us up. 
		if (doubleBuffered) *(ap++) = AGL_DOUBLEBUFFER;
		*(ap++) = AGL_NO_RECOVERY;
		*(ap++) = AGL_ACCELERATED;
		*(ap++) = AGL_WINDOW;
		*(ap++) = 0;
		AGLPixelFormat pf = aglChoosePixelFormat(NULL, 0, attribs);
		assert(pf);
		aglCtx = aglCreateContext(pf, NULL);
		assert(aglCtx);
		aglSetCurrentContext(aglCtx);
	}

	//now to attach the context to the window ...
	{
		aglSetDrawable(aglCtx, (AGLDrawable)GetWindowPort(wnd));
#endif
		disp.resize(VEC2ELEM(initSize));
	}

	//event handling
	{
		EventTargetRef target = GetWindowEventTarget(wnd);
		
		//keyboard
		EventTypeSpec keyUpEvents[] = {
			{kEventClassKeyboard, kEventRawKeyUp},
		};
		InstallEventHandler(target, NewEventHandlerUPP(KeyUpEventHandler), GetEventTypeCount(keyUpEvents), keyUpEvents, (void*)wnd, NULL);
		
		EventTypeSpec keyDownEvents[] = {
			{kEventClassKeyboard, kEventRawKeyDown},
		};
		InstallEventHandler(target, NewEventHandlerUPP(KeyDownEventHandler), GetEventTypeCount(keyDownEvents), keyDownEvents, (void*)wnd, NULL);

		EventTypeSpec modifierEvents[] = {
			{kEventClassKeyboard, kEventRawKeyModifiersChanged},
		};
		InstallEventHandler(target, NewEventHandlerUPP(KeyboardModifierEventHandler), GetEventTypeCount(modifierEvents), modifierEvents, (void*)wnd, NULL);
		
		//mouse
		EventTypeSpec mouseEvents[] = {
			{kEventClassMouse, kEventMouseUp},
			{kEventClassMouse, kEventMouseDown},
			{kEventClassMouse, kEventMouseMoved},
			{kEventClassMouse, kEventMouseDragged},
		};
		InstallEventHandler(target, NewEventHandlerUPP(MouseEventHandler), GetEventTypeCount(mouseEvents), mouseEvents, (void*)wnd, NULL);
		
		EventTypeSpec mouseWheelEvents[] = {
			{kEventClassMouse, kEventMouseWheelMoved},
		};
		InstallEventHandler(target, NewEventHandlerUPP(MouseWheelEventHandler), GetEventTypeCount(mouseWheelEvents), mouseWheelEvents, (void*)wnd, NULL);
		
		EventTypeSpec windowEvents[] = {
			{kEventClassWindow, kEventWindowClose},
			{kEventClassWindow, kEventWindowBoundsChanged},
		};
		InstallEventHandler(target, NewEventHandlerUPP(WindowEventHandler), GetEventTypeCount(windowEvents), windowEvents, (void*)wnd, NULL);
	}

	initKeyMaps();
}

void OSXSystem::shutdown() {
	//shutdown any objects *HERE*
#ifdef COCOA_INSTEAD_OF_CARBON
	cocoaShutdown();
#else
	aglDestroyContext(aglCtx);
	ReleaseWindow(wnd);
#endif
	//kill the event handlers if it means that much.  modular reasons and what not.  whatever.
	
	memSystemShutdown();

	//and reset...
	reset();
}

//backwards compatibility
void OSXSystem::requestExit() {
	exitRequested = true;
}

void OSXSystem::loop(void (*lc)()) {
	//store the main loop's callback method
	loopCallback = lc;
	
#if 0
	RunApplicationEventLoop();
#else
	//run the main loop
	//RunCurrentEventLoop(-log(0));	//infinity please
	EventTargetRef target = GetEventDispatcherTarget();
	EventRef event;
	while (!exitRequested) {
		//does this stall?
		if (ReceiveNextEvent(0, NULL, 0, true, &event) == noErr) {	//the duration (3rd param) was kEventDurationForever.  will putting this at 0 still let us grab events?
			SendEventToEventTarget(event, target);
			ReleaseEvent(event);
		}
		doLoop();
	}
#endif
}

void OSXSystem::doLoop() {
	preUpdate();
	if (loopCallback) loopCallback();
	postUpdate();
}

void OSXSystem::preUpdate() {
	//system update init
	disp.update();	//update resolution
	mouse.update();
	kbd.preUpdate();
}

void OSXSystem::postUpdate() {
	//system update shutdown
	kbd.postUpdate();
	//NOTICE - does this apply to all, or only to the local viewport?
	updateDisplayBuffers();
}

void OSXSystem::updateDisplayBuffers() {
	if (doubleBuffered) {
		//swap buffers in carbon?
#ifdef COCOA_INSTEAD_OF_CARBON
		CGLFlushDrawable(cglCtx);
#else
		aglSwapBuffers(aglCtx);	//does thsi only work if agl created the context?
#endif
	} else {
		glFlush();
	}
}
